% SPM
% Written by Romina Kazemivala

%load Parameters
clc;
clear all;
close all;
tic

%-------------------------------------------------------------------------
%%                          INPUT 
%-------------------------------------------------------------------------
%load Parameters
%CO, OUT, and RO Scattering Parameters
%1-3 of appendix C

rng = 1:41;%number of frequency samples
ftable = 8e9:1e8:12e9;
ftable = ftable(rng);

load S12_CO.mat
load S12_RO.mat
load S12_OUT_emptybear.mat
load S21_CO.mat
load S21_RO.mat
load S21_OUT_emptybear.mat


%Only the transmitted datasets
s_CO = {S21_CO};
s_RO = {S21_RO};
s_OUT = {S21_OUT};


zz = 150e-3;
zplane=1;
%==========================================================================
%%                    Define Variables
%==========================================================================
c0 = 2.998e8;

[nx, ny, nf,nz]=size(s_CO{1}); %nx, ny and nz are number of samples along x,y and z; fnum is number of frequency samples
 ns = size(s_CO);
 frng = 1:nf; %Change this if you want to see less data
 % Relative permittivity of background medium
 eps_RO = 1;
 
 % Relative permittivity of scattering probe in CO
 eps_CO = 12;
 
 % Volume of the scattering probe 
 r=2.5e-3;
 h=10e-3;
 vol_sp = pi*h*r^2;
 % CO_area = 5e-3^2;
 COsize=vol_sp; %in mm
 
 xstep = 5e-3;
 xrng = xstep*(nx-1);
 x=linspace(-xrng/2,xrng/2,nx);
 y=linspace(-xrng/2,xrng/2,ny);
 scan_step=xstep; %spatial sampling step in mm
 ROcontr=eps_RO; %relative permittivity of background (RO)
 COcontr=eps_CO-eps_RO;% the relative permittivity of the scattering probe and that of the RO)
 
 % Choose type of forward model: Born or Rytov
 approx=0; % Born is 1, Rytov is 0. 

 %Apodization filter
 apod=1;
 %display apodization filter
 show_apod=0;
 % enable low-pass filter setting lpf=1
 lpf=1;
 %display low-pass filter
 show_lpf=0;
 %focusing filter
 focusing=0;
 show_focus=0;
 %frequency scaling at the data stage
 freq_scaling=1;
 
 
 % if it is a 2D image, zstep is the distance between antennas.
 if length(zz)>1
     zstep = zz(2)-zz(1);
 else
     zstep = zz(1);
 end
  units = 'm';
 scaling_factor=(COsize/(scan_step^2*zstep))*COcontr; %calculating scaling factor for the cubic scattering probe. See Eq.(4.102)
%==========================================================================
 %%                          s_RO
%========================================================================== 
 if exist('s_RO_forCO','var') == 0
   disp('The RO dataset is the same for both the CO and OUT');
   s_RO_forCO = s_RO;
else
   disp('The RO dataset is independent for both the CO and OUT');
 end
 %-------------------------------------------------------------------------
 % Fill out s_RO if it contains only 1 value (i.e. size(s_RO{1} returns (1,1,nz,nf))
 %-------------------------------------------------------------------------
if size(s_RO{1}(:,1,1,1)) == 1
    for s = 1:ns
        for f = 1:nf
        s_RO_new{s}(:,:,f) =  (squeeze(s_RO{s}(1,1,2,f))*ones(nx,ny));
        end
    end
    s_RO = s_RO_new;

end
%==========================================================================
%%                   Apodization Filter
%==========================================================================
       %This part needs to be applied to the data before Fourier Transform 
       %to diminish cross-like artifacts.
     
if apod==1
 
    %%%%%%%%%%%%%%%%%%%%%  TUKEY FILTER  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    tukeytol = 0.6;
    tukey = tukeywin(nx,tukeytol)*tukeywin(nx,tukeytol)';
    apod_filter = tukey;
else
    %if we dont want an apodization filter, just multiply everything by 1!
    apod_filter = ones(nx);
end

%show the apodization filter
if show_apod
    figure;
    imagesc(apod_filter);
    title('Apodization Filter (Tukey Filter 0.6)');
end
  
%==========================================================================
%%       Obtain Real Domain Datasets Using Different Approximations
%==========================================================================

for s = 1:ns
    %----------------------------------------------------------------------
    %                     BORN Approximation
    %----------------------------------------------------------------------
    %Subtract RO from CO, OUT
    
    if approx
                

         %loop through each frequency
        for f = 1:nf
            s_OUT_scat{s}(:,:,f) = s_OUT{s}(:,:,f) - s_RO{s}(:,:,f,1);
            
            %loop through each z-plane
            for z = 1:nz
                s_CO_scat{s}(:,:,z,f) = s_CO{s}(:,:,f,z) - s_RO_forCO{s}(:,:,f);
                
            end
        end
                    disp('Born approximation');

    %----------------------------------------------------------------------
    %                    RYTOV Approximation
    %----------------------------------------------------------------------

    %Scattered Field OUT or CO = s_RO * ln (s_(OUT or CO)/s_RO)
    else
                        disp('Rytov approximation');

        s_OUT_scat{s} = s_RO{s}.*log(s_OUT{s}./s_RO{s});
        for z = 1:nz
            s_CO_scat{s}(:,:,z,:) = s_RO{s}.*log(squeeze(s_CO{s}(:,:,:,z))./s_RO_forCO{s});
        end
    end
     
     end


%==========================================================================
%%                        Fourier Transform
%==========================================================================
%Note that both frequency scaling and apodization filter pre-processing 
%steps are applied prior to the 2D FFT procedure
for s = 1:ns
    
    for f = 1:nf
        
        if freq_scaling %If the frequency normalization is applied on the extracted data
        %prior to the Fourier transform
        maxi = max(max(max(abs(s_CO{s}(:,:,f,:)))));
        else
            maxi=1;
        end
        
         s_OUT_fft{s}(:,:,f) = fft2(s_OUT_scat{s}(:,:,f).*apod_filter/maxi);

        for z = 1:nz
            s_CO_fft{s}(:,:,z,f) = fft2(s_CO_scat{s}(:,:,z,f).*apod_filter/maxi);

        end
        end
    end


%--------------------------------------------------------------------------
%%                   Scattering Probe (SP) Power Maps
%--------------------------------------------------------------------------
 % line (9) in Image Reconstruction section of Appendix C

% The CO Maps need to be Nz by Nz at every kx, ky position. Structing the
% loops to achieve this

%First we sum across all S-parameters, performing the conjugate
%multiplication
M_CO_nonorm = zeros(nx,ny,nz,nz,nf);

%for CO we need to loop through z twice, once for the plane being measured (zp),
%and once for the PSF plane using to try and measure that particular plane
%(zn)
for zp = 1:nz
    for zn = 1:nz
        for f=frng
            % sum the conjugates across all S-parameters
           for s = 1:ns
              %See Equation (41) on Page 16 of the Report
               M_CO_nonorm(:,:,zp,zn,f) = M_CO_nonorm(:,:,zp,zn,f) + ...
                   s_CO_fft{s}(:,:,zn,f).*conj(s_CO_fft{s}(:,:,zp,f)); 

           end
        end
    end
end
M_CO = sum(M_CO_nonorm,5)/nf;


%--------------------------------------------------------------------------
%%                     Generate OUT Power Maps
%--------------------------------------------------------------------------

%The OUT Maps need to be Nz by 1 at every kx, ky position. Structing the
%loops to achieve this
M_OUT_nonorm = zeros(nx,ny,nz,nf);
M_OUT_norm = zeros(nx,ny,nz);

%OUT only has one plane being measured, so we loop through all PSFs only
for zn = 1:nz
    %we need to normalize at every frequency
    for f=frng
        % sum the conjugates across all S-parameters
        for s = 1:ns
           
            %See Equation (40) on Page 16 of the Report

            M_OUT_nonorm(:,:,zn,f) = M_OUT_nonorm(:,:,zn,f) + ...
                s_OUT_fft{s}(:,:,f).*conj(s_CO_fft{s}(:,:,zn,f));
            
        end
                M_OUT_norm(:,:,zn) = M_OUT_norm(:,:,zn)+ M_OUT_nonorm(:,:,zn,f);

    end
end

% After all the summation we divide by the number of frequencies
M_OUT = M_OUT_norm/nf;
    
%--------------------------------------------------------------------------
%%                    LOW-PASS FILTER
%--------------------------------------------------------------------------


%%%%%%%%%%%%%%%%%%% Butterworth Filter %%%%%%%%%%%%%%%%%%%%%%%%%%%%%
if lpf == 1
   
%     calculate minimum wavelength
    lambda_min = c0/sqrt(abs((eps_RO)))/ftable(nf);
    
%     Calculate the maximum angle the antenna can see. This assumes that the
%     directivity of the antenna does not affect the maximum angle, which is
%     typically not the case in experimental reconstructions
    alphaa = atan(sqrt(2*(xrng/7)^2)/zz(nz));
    
%     calculate maximum kmax, and determine is location in the physical
%     FFT images
    kmax = 4*pi*sin(alphaa)/lambda_min;
    kxx = pi/xstep;
    dkx = kxx/nx;
    kmax_steps = kmax/dkx;
    dev =kmax_steps;
    
    
%     set the order of the Butterworth filter
    order=4;
    bfilt = ones(nx,ny);
%     BUTTERWORTH FILTER
    for i = 1:nx
        for j =1:ny
            chg = ((i-((nx-1)/2+1))^2 + (j-((ny-1)/2+1))^2)^.5;
%             Create Butterworth filter.
            bfilt(i,j)= 1/(1 + (chg/dev)^(2*order));
        end
    end
%     Shift the filter to align with the FFT versions of the CO and OUT
    lpf = fftshift(bfilt);
    
%     Any values below 0.01 are automatically set to 0. This reduces
%     computational costs
    lpf = lpf.*(lpf>0.01);
    
%         TESTING: Make the LPF scale so that the overal summation of the filter
%     remains the same as if there was no window (i.e. try to move the
%     overal signal power from the erroneous high frequnecy responses to the
%     lower frequency values
    
    lpf = lpf/sum(sum(lpf))*nx*ny;
else
   % if we dont want to filter, just multiply everything by 1!
    lpf = ones(nx);
end


 %show the filter
if show_lpf
    figure;
    imagesc(ifftshift(lpf));
    title('LPF (Butterworth Filter)');
    colorbar
    colormap parula
end


%--------------------------------------------------------------------------
%%                   INVERSE PROBLEM
%--------------------------------------------------------------------------
%Solving a linear system of equations in Equation (42) on Page 16
contrast_func_fft = zeros(nx,ny,nz);
for kx=1:nx
    for ky=1:ny
        %Quantitative Contrast Reconstruction in Fourier Space Step 11 in
        %Apendix C
        contrast_func_fft(kx,ky,:)=pinv(squeeze(M_CO(kx,ky,:,:)))*squeeze(M_OUT(kx,ky,:)); 
    end
end

%Once we have the contrast function (in fourier domain) we can apply our
%multiplication factors and generate the permittivity map (eps_out)
eps_out=zeros(nx,ny,nz); 
for z=1:nz
    %Obtain  Real-space  Contrast  via  2D  Inverse  Fourier  Transform
    %Step 13 in Appendix C
    %Note that LPF is applied to the contrast function in the Fourier space
    eps_out(:,:,z)=scaling_factor.*COcontr.*ifftshift(ifft2(lpf.*contrast_func_fft(:,:,z)))+ROcontr; 
end


%--------------------------------------------------------------------------
%%                      Plotting the results 
%--------------------------------------------------------------------------
%Final images are generated here.


for iz = 1:nz

       dx = xstep;
        
       X = (0:dx:((nx-1)*dx)/2).*1000/5;     
    %same as above: This image being printed is the total combined image
    %using all the s-parameters provided.

    fig = figure;

    subplot(1,3,1);
    imagesc(X,X,abs((rot90(eps_out(:,:,iz)))));
    title('Magnitude');
    colorbar;
    ystr = '{\ity} (cm)';
    xstr = '{\itx} (cm)';
    ylabel(ystr);
    xlabel(xstr);
    set(gca,'ydir','normal');

    subplot(1,3,2);
    imagesc(X,X,real(rot90((eps_out(:,:,iz)))));
    title('Real');
    set(gca,'ydir','normal');
    colorbar;
    ystr = '{\ity} (cm)';
    xstr = '{\itx} (cm)';
    ylabel(ystr);
    xlabel(xstr);
    
    subplot(1,3,3);
    imagesc(X,X,imag((rot90(eps_out(:,:,iz)))));
    title('Imaginary');
    colorbar;
    set(gca,'ydir','normal');
    ystr = '{\ity} (cm)';
    xstr = '{\itx} (cm)';
    ylabel(ystr);
    xlabel(xstr);
    
    set(findall(fig,'-property','FontSize'),'FontSize',15)
    set(fig, 'Position', [10, 10, 1200, 300]);

end

 
toc

